"""Create the training/validation datapipe for training the national MetNet/-2 Model"""
import datetime
import logging
from pathlib import Path
from typing import Union

import xarray
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.select import LocationPicker
from ocf_datapipes.training.common import (
    add_selected_time_slices_from_datapipes,
    get_and_return_overlapping_time_periods_and_t0,
    open_and_return_datapipes,
)
from ocf_datapipes.transform.xarray import StackXarray
from ocf_datapipes.utils.consts import NEW_NWP_MEAN, NEW_NWP_STD, RSS_MEAN, RSS_STD
from ocf_datapipes.utils.future import ThreadPoolMapperIterDataPipe as ThreadPoolMapper

xarray.set_options(keep_attrs=True)
logger = logging.getLogger("pseudo_irradiance_datapipe")
logger.setLevel(logging.DEBUG)


def normalize_pv(x):  # So it can be pickled
    """
    Normalize the GSP data

    Args:
        x: Input DataArray

    Returns:
        Normalized DataArray
    """
    return x / x.capacity_watt_power


def _remove_nans(x):
    return x.fillna(0.0)


def _get_numpy_from_xarray(x):
    return x.to_numpy()


def _load_xarray_values(x):
    return x.load()


def pseudo_irradiance_datapipe(
    configuration_filename: Union[Path, str],
    use_sun: bool = True,
    use_nwp: bool = True,
    use_sat: bool = True,
    use_hrv: bool = True,
    use_pv: bool = True,
    use_topo: bool = True,
    use_future: bool = False,
    size: int = 256,
    start_time: datetime.datetime = datetime.datetime(2014, 1, 1),
    end_time: datetime.datetime = datetime.datetime(2023, 1, 1),
    batch_size: int = 1,
) -> IterDataPipe:
    """
    Make Pseudo-Irradience Datapipe

    This outputs PV generation

    Args:
        configuration_filename: the configruation filename for the pipe
        use_sun: Whether to add sun features or not
        use_pv: Whether to use PV input or not
        use_hrv: Whether to use HRV Satellite or not
        use_sat: Whether to use non-HRV Satellite or not
        use_nwp: Whether to use NWP or not
        use_topo: Whether to use topographic map or not
        start_time: Start time to select on
        end_time: End time to select from
        use_future: Whether to use future frames as well (for labelling)
        size: Size, in pixels, of the output image
        batch_size: Batch size for the datapipe

    Returns: datapipe
    """

    # load datasets
    used_datapipes = open_and_return_datapipes(
        configuration_filename=configuration_filename,
        use_nwp=use_nwp,
        use_topo=use_topo,
        use_sat=use_sat,
        use_hrv=use_hrv,
        use_gsp=False,
        use_pv=use_pv,
    )
    # Load GSP national data
    used_datapipes["pv"] = used_datapipes["pv"].select_train_test_time(start_time, end_time)

    # Now get overlapping time periods
    used_datapipes = get_and_return_overlapping_time_periods_and_t0(used_datapipes, key_for_t0="pv")

    # And now get time slices
    used_datapipes = add_selected_time_slices_from_datapipes(used_datapipes)

    # Now do the extra processing
    pv_history = used_datapipes["pv"].normalize(normalize_fn=normalize_pv)
    pv_datapipe = used_datapipes["pv_future"].normalize(normalize_fn=normalize_pv)
    # Split into GSP for target, only national, and one for history
    pv_datapipe, pv_loc_datapipe = pv_datapipe.fork(2)
    pv_loc_datapipe, pv_id_datapipe = LocationPicker(pv_loc_datapipe).fork(2)

    if "nwp" in used_datapipes.keys():
        # take nwp time slices
        logger.debug("Take NWP time slices")
        nwp_datapipe = used_datapipes["nwp"].normalize(mean=NEW_NWP_MEAN, std=NEW_NWP_STD)
        pv_loc_datapipe, pv_nwp_image_loc_datapipe = pv_loc_datapipe.fork(2)
        nwp_datapipe = nwp_datapipe.select_spatial_slice_pixels(
            pv_nwp_image_loc_datapipe,
            roi_height_pixels=size,
            roi_width_pixels=size,
            x_dim_name="x_osgb",
            y_dim_name="y_osgb",
        )
        nwp_datapipe = ThreadPoolMapper(
            nwp_datapipe, _load_xarray_values, max_workers=8, scheduled_tasks=batch_size
        )

    if "sat" in used_datapipes.keys():
        logger.debug("Take Satellite time slices")
        # take sat time slices
        sat_datapipe = used_datapipes["sat"].normalize(mean=RSS_MEAN, std=RSS_STD)
        pv_loc_datapipe, pv_sat_image_loc_datapipe = pv_loc_datapipe.fork(2)
        sat_datapipe = sat_datapipe.select_spatial_slice_pixels(
            pv_sat_image_loc_datapipe,
            roi_height_pixels=size,
            roi_width_pixels=size,
            x_dim_name="x_geostationary",
            y_dim_name="y_geostationary",
        )
        sat_datapipe = ThreadPoolMapper(
            sat_datapipe, _load_xarray_values, max_workers=8, scheduled_tasks=batch_size
        )

    if "hrv" in used_datapipes.keys():
        logger.debug("Take HRV Satellite time slices")
        sat_hrv_datapipe = used_datapipes["hrv"].normalize(mean=RSS_MEAN, std=RSS_STD)
        pv_loc_datapipe, pv_hrv_image_loc_datapipe = pv_loc_datapipe.fork(2)
        sat_hrv_datapipe = sat_hrv_datapipe.select_spatial_slice_pixels(
            pv_hrv_image_loc_datapipe,
            roi_height_pixels=size,
            roi_width_pixels=size,
            x_dim_name="x_geostationary",
            y_dim_name="y_geostationary",
        )
        sat_hrv_datapipe = ThreadPoolMapper(
            sat_hrv_datapipe, _load_xarray_values, max_workers=8, scheduled_tasks=batch_size
        )

    if "topo" in used_datapipes.keys():
        topo_datapipe = used_datapipes["topo"].map(_remove_nans)
        pv_loc_datapipe, pv_hrv_image_loc_datapipe = pv_loc_datapipe.fork(2)
        topo_datapipe = topo_datapipe.select_spatial_slice_pixels(
            pv_hrv_image_loc_datapipe,
            roi_height_pixels=size,
            roi_width_pixels=size,
            x_dim_name="x_osgb",
            y_dim_name="y_osgb",
        )
        topo_datapipe = ThreadPoolMapper(
            topo_datapipe, _load_xarray_values, max_workers=8, scheduled_tasks=batch_size
        )
    # Setting seed in these to keep them the same for creating image and metadata
    if "hrv" in used_datapipes.keys():
        sat_hrv_datapipe, sat_gsp_datapipe, sat_meta_datapipe = sat_hrv_datapipe.fork(3)
        pv_history, pv_meta = pv_history.fork(2)
        pv_history = pv_history.create_pv_image(image_datapipe=sat_gsp_datapipe, seed=1337)
        pv_meta = pv_meta.create_pv_metadata_image(image_datapipe=sat_meta_datapipe, seed=1337)
    elif "sat" in used_datapipes.keys():
        sat_datapipe, sat_gsp_datapipe, sat_meta_datapipe = sat_datapipe.fork(3)
        pv_history, pv_meta = pv_history.fork(2)
        pv_history = pv_history.create_pv_image(image_datapipe=sat_gsp_datapipe, seed=1337)
        pv_meta = pv_meta.create_pv_metadata_image(image_datapipe=sat_meta_datapipe, seed=1337)
    elif "nwp" in used_datapipes.keys():
        nwp_datapipe, nwp_gsp_datapipe, nwp_meta_datapipe = nwp_datapipe.fork(3)
        pv_history, pv_meta = pv_history.fork(2)
        pv_history = pv_history.create_pv_image(
            image_datapipe=nwp_gsp_datapipe, image_dim="osgb", seed=1337
        )
        pv_meta = pv_meta.create_pv_metadata_image(
            image_datapipe=nwp_meta_datapipe, image_dim="osgb", seed=1337
        )

    # Need to have future in image as well
    if "hrv" in used_datapipes.keys():
        sat_hrv_datapipe, sat_future_datapipe = sat_hrv_datapipe.fork(2)
        pv_datapipe = pv_datapipe.create_pv_image(image_datapipe=sat_future_datapipe)
    elif "sat" in used_datapipes.keys():
        sat_datapipe, sat_future_datapipe = sat_datapipe.fork(2)
        pv_datapipe = pv_datapipe.create_pv_image(image_datapipe=sat_future_datapipe)
    elif "nwp" in used_datapipes.keys():
        nwp_datapipe, nwp_future_datapipe = nwp_datapipe.fork(2)
        pv_datapipe = pv_datapipe.create_pv_image(
            image_datapipe=nwp_future_datapipe, image_dim="osgb"
        )

    if use_sun:
        if "nwp" in used_datapipes.keys():
            nwp_datapipe, sun_image_datapipe = nwp_datapipe.fork(2)
            sun_image_datapipe = sun_image_datapipe.create_sun_image(
                normalize=True, image_dim="osgb", time_dim="target_time_utc"
            )
        elif "hrv" in used_datapipes.keys():
            # Want it at highest resolution possible
            sat_hrv_datapipe, sun_image_datapipe = sat_hrv_datapipe.fork(2)
            sun_image_datapipe = sun_image_datapipe.create_sun_image(normalize=True)
        elif "sat" in used_datapipes.keys():
            sat_datapipe, sun_image_datapipe = sat_datapipe.fork(2)
            sun_image_datapipe = sun_image_datapipe.create_sun_image(normalize=True)

    if "nwp" in used_datapipes.keys():
        nwp_datapipe, time_image_datapipe = nwp_datapipe.fork(2)
        time_image_datapipe = time_image_datapipe.create_time_image(
            image_dim="osgb", time_dim="target_time_utc"
        )
    elif "hrv" in used_datapipes.keys():
        # Want it at highest resolution possible
        sat_hrv_datapipe, time_image_datapipe = sat_hrv_datapipe.fork(2)
        time_image_datapipe = time_image_datapipe.create_time_image()
    elif "sat" in used_datapipes.keys():
        sat_datapipe, time_image_datapipe = sat_datapipe.fork(2)
        time_image_datapipe = time_image_datapipe.create_time_image()
    else:
        time_image_datapipe = None

    modalities = []
    modalities.append(pv_history)
    if "nwp" in used_datapipes.keys():
        modalities.append(nwp_datapipe)
    if "hrv" in used_datapipes.keys():
        modalities.append(sat_hrv_datapipe)
    if "sat" in used_datapipes.keys():
        modalities.append(sat_datapipe)
    if "topo" in used_datapipes.keys():
        modalities.append(topo_datapipe)
    if use_sun:
        modalities.append(sun_image_datapipe)
    if time_image_datapipe is not None:
        modalities.append(time_image_datapipe)

    stacked_xarray_inputs = StackXarray(modalities)

    pv_datapipe = pv_datapipe.map(_get_numpy_from_xarray)
    pv_meta = pv_meta.map(_get_numpy_from_xarray)
    return stacked_xarray_inputs.batch(batch_size).zip_ocf(
        pv_meta.batch(batch_size), pv_datapipe.batch(batch_size)
    )  # Makes (Inputs, Label) tuples



"""Select overlapping time slices for training"""
import logging
from typing import Iterable, Optional

import pandas as pd
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("select_overlapping_time_slice")
class SelectOverlappingTimeSliceIterDataPipe(IterDataPipe):
    """Source DataPipes are from the contiguous_time_period"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        secondary_datapipes: Iterable[IterDataPipe],
        location_datapipe: Optional[IterDataPipe] = None,
    ):
        """
        Source DataPipes are from the contiguous_time_period

        Args:
            source_datapipe: First Datapipe to compute the intersection of
            secondary_datapipes: More Datapipe to compute the intersection of
            location_datapipe: location fo data pipe. This allows quick loading of
                time periods if present
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.secondary_datapipes = secondary_datapipes
        self.location_datapipe = location_datapipe
        self.time_periods_id = {}

    def __iter__(self) -> pd.DataFrame:
        if self.location_datapipe is None:
            for set_of_pd_datas in self.source_datapipe.zip_ocf(*self.secondary_datapipes):
                time_periods = intersection_of_multiple_dataframes_of_periods(list(set_of_pd_datas))

                logger.debug(f"Found {len(time_periods)} time periods")
                assert len(time_periods) > 0

                yield time_periods
        else:
            for set_of_pd_datas in self.source_datapipe.zip_ocf(
                *self.secondary_datapipes, self.location_datapipe
            ):
                location = set_of_pd_datas[-1]
                set_of_pd_datas = set_of_pd_datas[:-1]

                id = int(location.id)

                if id in self.time_periods_id.keys():
                    logger.debug(f"Time periods for id {id} from store")
                    logger.debug(len(self.time_periods_id[id]))
                    time_periods = self.time_periods_id[id]
                else:
                    logger.debug(f"Time periods for id {id} not in store")
                    time_periods = intersection_of_multiple_dataframes_of_periods(
                        list(set_of_pd_datas)
                    )

                    logger.debug(f"Found {len(time_periods)} time periods")
                    assert len(time_periods) > 0

                    self.time_periods_id[id] = time_periods
                    logger.debug(len(time_periods))

                yield time_periods


def intersection_of_multiple_dataframes_of_periods(
    time_periods: list[pd.DataFrame],
) -> pd.DataFrame:
    """Find the intersection of a list of time periods.

    See the docstring of intersection_of_2_dataframes_of_periods() for more details.
    """
    assert len(time_periods) > 0
    if len(time_periods) == 1:
        return time_periods[0]
    intersection = intersection_of_2_dataframes_of_periods(time_periods[0], time_periods[1])
    for time_period in time_periods[2:]:
        intersection = intersection_of_2_dataframes_of_periods(intersection, time_period)
    return intersection


def intersection_of_2_dataframes_of_periods(a: pd.DataFrame, b: pd.DataFrame) -> pd.DataFrame:
    """Find the intersection of two pd.DataFrames of time periods.

    Each row of each pd.DataFrame represents a single time period.  Each pd.DataFrame has
    two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').

    A typical use-case is that each pd.DataFrame represents all the time periods where
    a `DataSource` has contiguous, valid data.

    Here's a graphical example of two pd.DataFrames of time periods and their intersection:

                 ----------------------> TIME ->---------------------
               a: |-----|   |----|     |----------|     |-----------|
               b:    |--------|                       |----|    |---|
    intersection:    |--|   |-|                         |--|    |---|

    Args:
        a: pd.DataFrame where each row represents a time period.  The pd.DataFrame has
        two columns: start_dt and end_dt.
        b: pd.DataFrame where each row represents a time period.  The pd.DataFrame has
        two columns: start_dt and end_dt.

    Returns:
        Sorted list of intersecting time periods represented as a pd.DataFrame with two columns:
        start_dt and end_dt.
    """
    if a.empty or b.empty:
        return pd.DataFrame(columns=["start_dt", "end_dt"])

    all_intersecting_periods = []
    for a_period in a.itertuples():
        # Five ways in which two periods may overlap:
        # a: |----| or |---|   or  |---| or   |--|   or |-|
        # b:  |--|       |---|   |---|      |------|    |-|
        # In all five, `a` must always start before `b` ends,
        # and `a` must always end after `b` starts:
        overlapping_periods = b[(a_period.start_dt < b.end_dt) & (a_period.end_dt > b.start_dt)]

        # There are two ways in which two periods may *not* overlap:
        # a: |---|        or        |---|
        # b:       |---|      |---|
        # `overlapping` will not include periods which do *not* overlap.

        # Now find the intersection of each period in `overlapping_periods` with
        # the period from `a` that starts at `a_start_dt` and ends at `a_end_dt`.
        # We do this by clipping each row of `overlapping_periods`
        # to start no earlier than `a_start_dt`, and end no later than `a_end_dt`.

        # First, make a copy, so we don't clip the underlying data in `b`.
        intersecting_periods = overlapping_periods.copy()
        intersecting_periods.start_dt.clip(lower=a_period.start_dt, inplace=True)
        intersecting_periods.end_dt.clip(upper=a_period.end_dt, inplace=True)

        all_intersecting_periods.append(intersecting_periods)

    all_intersecting_periods = pd.concat(all_intersecting_periods)
    return all_intersecting_periods.sort_values(by="start_dt").reset_index(drop=True)



"""Preprocessing for MetNet-type inputs"""
import itertools
from typing import List

import numpy as np
import pvlib
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils import Zipper
from ocf_datapipes.utils.geospatial import (
    load_geostationary_area_definition_and_transform_osgb,
    load_geostationary_area_definition_and_transform_to_latlon,
    osgb_to_lat_lon,
)
from ocf_datapipes.utils.parallel import run_with_threadpool
from ocf_datapipes.utils.utils import trigonometric_datetime_transformation

ELEVATION_MEAN = 37.4
ELEVATION_STD = 12.7
AZIMUTH_MEAN = 177.7
AZIMUTH_STD = 41.7


@functional_datapipe("preprocess_metnet")
class PreProcessMetNetIterDataPipe(IterDataPipe):
    """Preprocess set of Xarray datasets similar to MetNet-1"""

    def __init__(
        self,
        source_datapipes: List[IterDataPipe],
        location_datapipe: IterDataPipe,
        context_width: int,
        context_height: int,
        center_width: int,
        center_height: int,
        output_height_pixels: int,
        output_width_pixels: int,
        add_sun_features: bool = False,
        only_sun: bool = False,
    ):
        """

        Processes set of Xarray datasets similar to MetNet

        In terms of taking all available source datapipes:
        1. selecting the same context area of interest
        2. Creating a center crop of the center_height, center_width
        3. Downsampling the context area of interest to the same shape as the center crop
        4. Stacking those context images on the center crop.
        5. Add Month, Day, Hour channels for each input time
        6. Add Sun position as well?

        This would be designed originally for NWP+Satellite+Topographic data sources.
        To add the PV power for lots of sites, the PV power would
        need to be able to be on a grid for the context/center
        crops and then for the downsample

        This also appends Lat/Lon coordinates to the stack,
         and returns a new Numpy array with the stacked data

        Args:
            source_datapipes: Datapipes that emit xarray datasets
                with latitude/longitude coordinates included
            location_datapipe: Datapipe emitting location coordinate for center of example
            context_width: Width of the context area
            context_height: Height of the context area
            center_width: Center width of the area of interest
            center_height: Center height of the area of interest
            output_height_pixels: Output height in pixels
            output_width_pixels: Output width in pixels
            add_sun_features: Whether to calculate and
            add Sun elevation and azimuth for each center pixel
            only_sun: Whether to only output sun features
                Assumes only one input to give the coordinates
        """
        self.source_datapipes = source_datapipes
        self.location_datapipe = location_datapipe
        self.context_width = context_width
        self.context_height = context_height
        self.center_width = center_width
        self.center_height = center_height
        self.output_height_pixels = output_height_pixels
        self.output_width_pixels = output_width_pixels
        self.add_sun_features = add_sun_features
        self.only_sun = only_sun

    def __iter__(self) -> np.ndarray:
        for xr_datas, location in Zipper(Zipper(*self.source_datapipes), self.location_datapipe):
            # TODO Use the Lat/Long coordinates of the center array for the lat/lon stuff
            # Do the resampling and cropping in parallel
            xr_datas = run_with_threadpool(
                zip(
                    _bicycle(xr_datas),
                    itertools.repeat(location),
                    itertools.chain.from_iterable(
                        zip(
                            itertools.repeat(self.center_width),
                            itertools.repeat(self.context_width),
                        )
                    ),
                    itertools.chain.from_iterable(
                        zip(
                            itertools.repeat(self.center_height),
                            itertools.repeat(self.context_height),
                        )
                    ),
                    itertools.repeat(self.output_height_pixels),
                    itertools.repeat(self.output_width_pixels),
                ),
                _crop_and_resample_wrapper,
                max_workers=8,
                scheduled_tasks=int(len(xr_datas) * 2),  # One for center, one for context
            )
            xr_datas = list(xr_datas)
            # Output is then list of center, context, center, context, etc.
            # So we need to split the list into two lists of the same length,
            # one with centers, one with contexts
            centers = xr_datas[::2]
            contexts = xr_datas[1::2]
            # Now do the first one for the sun and other features
            xr_center = centers[0]
            _extra_time_dim = (
                "target_time_utc" if "target_time_utc" in xr_center.dims else "time_utc"
            )
            # Add in time features for each timestep
            time_image = _create_time_image(
                xr_center,
                time_dim=_extra_time_dim,
                output_height_pixels=self.output_height_pixels,
                output_width_pixels=self.output_width_pixels,
            )
            contexts.append(time_image)
            # Need to add sun features
            if self.add_sun_features:
                sun_image = _create_sun_image(
                    image_xr=xr_center,
                    x_dim="x_osgb" if "x_osgb" in xr_center.dims else "x_geostationary",
                    y_dim="y_osgb" if "y_osgb" in xr_center.dims else "y_geostationary",
                    time_dim=_extra_time_dim,
                    normalize=True,
                )
                if self.only_sun:
                    contexts = [time_image, sun_image]
                else:
                    contexts.append(sun_image)
            for xr_index in range(len(centers)):
                xr_center = centers[xr_index]
                xr_context = contexts[xr_index]
                xr_center = xr_center.to_numpy()
                xr_context = xr_context.to_numpy()
                if len(xr_center.shape) == 2:  # Need to add channel dimension
                    xr_center = np.expand_dims(xr_center, axis=0)
                    xr_context = np.expand_dims(xr_context, axis=0)
                if len(xr_center.shape) == 3:  # Need to add channel dimension
                    xr_center = np.expand_dims(xr_center, axis=1)
                    xr_context = np.expand_dims(xr_context, axis=1)
                centers[xr_index] = xr_center
                contexts[xr_index] = xr_context
            # Pad out time dimension to be the same, using the largest one
            # All should have 4 dimensions at this point
            max_time_len = max(
                np.max([c.shape[0] for c in centers]), np.max([c.shape[0] for c in contexts])
            )
            for i in range(len(centers)):
                centers[i] = np.pad(
                    centers[i],
                    pad_width=(
                        (0, max_time_len - centers[i].shape[0]),
                        (0, 0),
                        (0, 0),
                        (0, 0),
                    ),
                    mode="constant",
                    constant_values=0.0,
                )
            for i in range(len(contexts)):
                contexts[i] = np.pad(
                    contexts[i],
                    pad_width=(
                        (0, max_time_len - contexts[i].shape[0]),
                        (0, 0),
                        (0, 0),
                        (0, 0),
                    ),
                    mode="constant",
                    constant_values=0.0,
                )
            stacked_data = np.concatenate([*centers, *contexts], axis=1)
            yield stacked_data


def _crop_and_resample_wrapper(args):
    return _crop_and_resample(*args)


def _bicycle(xr_datas):
    for xr_data in xr_datas:
        yield xr_data
        yield xr_data


def _crop_and_resample(
    xr_data: xr.Dataset,
    location,
    context_width,
    context_height,
    output_height_pixels,
    output_width_pixels,
):
    xr_context: xr.Dataset = _get_spatial_crop(
        xr_data,
        location=location,
        roi_width_meters=context_width,
        roi_height_meters=context_height,
    )
    # Resamples to the same number of pixels for both center and contexts
    xr_context = _resample_to_pixel_size(xr_context, output_height_pixels, output_width_pixels)
    return xr_context


def _get_spatial_crop(xr_data, location, roi_height_meters: int, roi_width_meters: int):
    # Compute the index for left and right:
    half_height = roi_height_meters // 2
    half_width = roi_width_meters // 2

    left = location.x - half_width
    right = location.x + half_width
    bottom = location.y - half_height
    top = location.y + half_height
    if "x_geostationary" in xr_data.coords:
        # Convert to geostationary edges
        _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
        left, bottom = _osgb_to_geostationary(xx=left, yy=bottom)
        right, top = _osgb_to_geostationary(xx=right, yy=top)
        x_mask = (left <= xr_data.x_geostationary) & (xr_data.x_geostationary <= right)
        y_mask = (xr_data.y_geostationary <= top) & (  # Y is flipped
            bottom <= xr_data.y_geostationary
        )
        selected = xr_data.isel(x_geostationary=x_mask, y_geostationary=y_mask)
    elif "x" in xr_data.coords:
        _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
        left, bottom = _osgb_to_geostationary(xx=left, yy=bottom)
        right, top = _osgb_to_geostationary(xx=right, yy=top)
        x_mask = (left <= xr_data.x) & (xr_data.x <= right)
        y_mask = (xr_data.y <= top) & (bottom <= xr_data.y)  # Y is flipped
        selected = xr_data.isel(x=x_mask, y=y_mask)
    else:
        # Select data in the region of interest:
        x_mask = (left <= xr_data.x_osgb) & (xr_data.x_osgb <= right)
        y_mask = (xr_data.y_osgb <= top) & (bottom <= xr_data.y_osgb)
        selected = xr_data.isel(x_osgb=x_mask, y_osgb=y_mask)

    return selected


def _resample_to_pixel_size(xr_data, height_pixels, width_pixels) -> np.ndarray:
    if "x_geostationary" in xr_data.dims:
        x_coords = xr_data["x_geostationary"].values
        y_coords = xr_data["y_geostationary"].values
    elif "x_osgb" in xr_data.dims:
        x_coords = xr_data["x_osgb"].values
        y_coords = xr_data["y_osgb"].values
    else:
        x_coords = xr_data["x"].values
        y_coords = xr_data["y"].values
    # Resample down to the number of pixels wanted
    x_coords = np.linspace(x_coords[0], x_coords[-1], num=width_pixels)
    y_coords = np.linspace(y_coords[0], y_coords[-1], num=height_pixels)
    if "x_geostationary" in xr_data.dims:
        xr_data = xr_data.interp(
            x_geostationary=x_coords, y_geostationary=y_coords, method="linear"
        )
    elif "x_osgb" in xr_data.dims:
        xr_data = xr_data.interp(x_osgb=x_coords, y_osgb=y_coords, method="linear")
    else:
        xr_data = xr_data.interp(x=x_coords, y=y_coords, method="linear")
    # Extract just the data now
    return xr_data


def _create_time_image(xr_data, time_dim: str, output_height_pixels: int, output_width_pixels: int):
    # Create trig decomposition of datetime values, tiled over output height and width
    datetimes = xr_data[time_dim].values
    trig_decomposition = trigonometric_datetime_transformation(datetimes)
    tiled_data = np.expand_dims(trig_decomposition, (2, 3))
    tiled_data = np.tile(tiled_data, (1, 1, output_height_pixels, output_width_pixels))
    return tiled_data


def _create_sun_image(image_xr, x_dim, y_dim, time_dim, normalize):
    # Create empty image to use for the PV Systems, assumes image has x and y coordinates
    sun_image = np.zeros(
        (
            2,  # Azimuth and elevation
            len(image_xr[y_dim]),
            len(image_xr[x_dim]),
            len(image_xr[time_dim]),
        ),
        dtype=np.float32,
    )
    if "geostationary" in x_dim:
        transform_to_latlon = load_geostationary_area_definition_and_transform_to_latlon(image_xr)
        lats, lons = transform_to_latlon(xx=image_xr[x_dim].values, yy=image_xr[y_dim].values)
    else:
        transform_to_latlon = osgb_to_lat_lon
        lats, lons = transform_to_latlon(x=image_xr.x_osgb.values, y=image_xr.y_osgb.values)
    time_utc = image_xr[time_dim].values

    # Loop round each example to get the Sun's elevation and azimuth:
    # Go through each time on its own, lat lons still in order of image
    # TODO Make this faster
    # dt = pd.DatetimeIndex(dt)  # pvlib expects a `pd.DatetimeIndex`.
    for example_idx, (lat, lon) in enumerate(zip(lats, lons)):
        solpos = pvlib.solarposition.get_solarposition(
            time=time_utc,
            latitude=lat,
            longitude=lon,
            # Which `method` to use?
            # pyephem seemed to be a good mix between speed and ease but causes segfaults!
            # nrel_numba doesn't work when using multiple worker processes.
            # nrel_c is probably fastest but requires C code to be manually compiled:
            # https://midcdmz.nrel.gov/spa/
        )
        sun_image[0][:][example_idx] = solpos["azimuth"]
        sun_image[1][example_idx][:] = solpos["elevation"]

    # Flip back to normal ordering
    sun_image = np.transpose(sun_image, [3, 0, 1, 2])

    # Normalize.
    if normalize:
        sun_image[:, 0] = (sun_image[:, 0] - AZIMUTH_MEAN) / AZIMUTH_STD
        sun_image[:, 1] = (sun_image[:, 1] - ELEVATION_MEAN) / ELEVATION_STD
    return sun_image



from ocf_datapipes.transform.xarray import StackXarray
from ocf_datapipes.select import LocationPicker
import numpy as np


def test_stack_xarray(sat_hrv_datapipe, sat_datapipe, nwp_datapipe, passiv_datapipe):
    loc1, loc2 = LocationPicker(passiv_datapipe).fork(2)
    sat_hrv_datapipe = sat_hrv_datapipe.select_spatial_slice_pixels(
        loc1,
        roi_height_pixels=32,
        roi_width_pixels=32,
        y_dim_name="y_geostationary",
        x_dim_name="x_geostationary",
    )
    sat_datapipe = sat_datapipe.select_spatial_slice_pixels(
        loc2,
        roi_height_pixels=32,
        roi_width_pixels=32,
        y_dim_name="y_geostationary",
        x_dim_name="x_geostationary",
    )
    datapipe = StackXarray([sat_hrv_datapipe, sat_datapipe])
    data = next(iter(datapipe))
    assert isinstance(data, np.ndarray)
    assert not np.isnan(data).any()
